#!/usr/bin/env python

"""
    Novel miRNA prediction
    Created Feb 2017
    Copyright (C) Damien Farrell

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 3
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
"""

from __future__ import absolute_import, print_function
import sys, os, string, types
import shutil, glob, collections
import itertools
from itertools import islice
import numpy as np
import pandas as pd
from . import base, utils

path = os.path.dirname(os.path.abspath(__file__))
datadir = os.path.join(path, 'data')
CLASSIFIER = None

def get_triplets(seq, struct):
    """Get triplet elements used by Xue et al."""

    tr=['(((', '((.', '(..', '(.(', '.((', '.(.', '..(' , '...']
    nuc = ['A','G','T','C']
    d={}
    for i in nuc:
        for j in tr:
            d[i+j]=0
    struct=struct.replace(')','(')
    l = len(seq)-len(seq)%3
    for i in range(0,l,3):
        n = seq[i+1]+struct[i:i+3]
        if n in d:
            d[n]+=1
    return d

def get_biggest_stem(bg):
    biggest_stem = (-1, 'x')
    for s in bg.stem_iterator():
        if bg.stem_length(s) > biggest_stem[0]:
            biggest_stem = (bg.stem_length(s), s)
    return biggest_stem

def get_stem_pairs(bg):
    """Stem pairs of rna hairpin"""

    pairs=[]
    for s in bg.stem_iterator():
        for p in bg.stem_bp_iterator(s):
            pairs.append( (bg.seq[p[0]-1],bg.seq[p[1]-1]) )
    return pairs

def get_stem_matches(bg):
    """Find rna stem mismatches"""

    pairs = get_stem_pairs(bg)
    wc = {'G':'C','C':'G','T':'A','A':'T'}
    matches = [True if i[1]==wc[i[0]] else False for i in pairs]
    return matches

def GC(seq):
    target_count = sum(1 for x in seq if x in ['G','C'])
    if len(seq) == 0:
        return 0
    return float(target_count) / len(seq) * 100

def build_rna_features(seq, struct=None, sc=None, mature=None):
    """Get features for mirna sequence"""

    import forgi.graph.bulge_graph as cgb

    if struct == None:
        struct,sc = utils.rnafold(seq)

    feats = {}
    #feats['reads'] = reads.reads.sum()
    feats['length'] = len(seq)
    feats['mfe'] = round(sc/len(seq),3)
    #print seq
    #print struct

    bg = utils.get_bg(seq, struct)
    try:
        h0seq = bg.get_define_seq_str('h0')[0]
        feats['loops'] = len(list(bg.hloop_iterator()))
        feats['loop_length'] = bg.get_bulge_dimensions('h0')[0]
    except:
        h0seq=''
        feats['loops'] = 0
        feats['loop_length'] = 0
    feats['gc'] = GC(seq)
    feats['loop_gc']= GC(h0seq)
    feats['stem_length'] = len(get_stem_pairs(bg))
    feats['longest_stem'] = get_biggest_stem(bg)[0]
    bulges = [bg.get_bulge_dimensions(i) for i in bg.iloop_iterator()]
    feats['bulges'] = len(bulges)
    try:
        feats['longest_bulge'] =  max(max(zip(*bulges)))
    except:
        feats['longest_bulge'] = 0
    bulgematches = [True if i[0]==i[1] else False for i in bulges]
    feats['bulges_symmetric'] = bulgematches.count(True)
    feats['bulges_asymmetric'] = bulgematches.count(False)
    sm = get_stem_matches(bg)
    feats['stem_mismatches'] = sm.count(False)

    if mature == None:
        #mature should be given - place holder for code to guess it later
        start = np.random.randint(1,len(sm)-20)
        end = start+22
    else:
        start = utils.find_subseq(seq, mature)
        end = start+len(mature)
        #print start, end
    feats['mature_mismatches'] = sm[start:end].count(False)
    feats['struct'] = struct
    tr = get_triplets(seq, struct)
    feats.update(tr)
    return feats

def find_star_sequence(seq, mature, struct=None):
    """Estimate the star sequence from a given mature and precursor."""

    bg = utils.get_bg(seq, struct)
    start = utils.find_subseq(seq, mature)+1
    end = start + len(mature)
    #print (start, end)
    #find end of mature to estimate start of star
    for p in bg.adjacent_stem_pairs_iterator():
        #print (p)
        for s in p:
            #print (s)
            x = list(bg.stem_bp_iterator(s))
            #print (x)
            for i in x:
                if start == i[0]:
                    #print ('found start', i )
                    star = seq[i[1]-len(mature):i[1]+2]
                    return star
                elif start == i[1]:
                    #print ('found start', i )
                    star = seq[i[0]-len(mature)+1:i[0]]
                    return star
    print ('star not found')
    return

def check_hairpin(seq, struct):
    """Remove free ends of hairpin"""

    bg = utils.get_bg(seq, struct)
    loops = list(bg.hloop_iterator())
    if len(loops) != 1:
        return seq, struct
    p = list(bg.stem_bp_iterator('s0'))[0]
    s = p[0]-1; e=p[1]
    return seq[s:e], struct[s:e]

def check_mature(seq, struct, mature):
    """Check if the mature sequence is not in hairpin loop and inside stem"""

    bg = utils.get_bg(seq, struct)
    start = utils.find_subseq(seq, mature)+1
    end = start + len(mature)
    loops = list(bg.hloop_iterator())
    if len(loops)==0:
        return 'too many loops'
    l = list(bg.define_range_iterator('h0'))[0]
    if (start<l[0] and end>l[0]) or (start<l[1] and end>l[1]):
        return 'mature in loop'
    p = list(bg.stem_bp_iterator('s0'))[0]
    #print (p[0], p[1], start, end)
    if start == 0:
        return 'mature not found'
    if end>p[1]+2:
        return '3p mature outside'
    return 'ok'

def get_positives(species='hsa'):
    """Get known mirbase hairpins for training precursor classifier. """

    reload(base)
    mirs = base.get_mirbase(species)
    feats=[]
    for i,row in mirs.iterrows():
        f = build_rna_features(row.precursor, mature=row.mature1_seq)
        f['seq'] = row.precursor
        f['mature'] = row.mature1_seq
        f['star'] = row.mature2_seq
        feats.append(f)

    result = pd.DataFrame(feats)
    result.to_csv('known_mirna_features.csv', index=False)
    return result

def get_negatives(fasta_file=None, samples=2000):
    """Create a negative pseudo mirna set for training classifier"""

    if fasta_file == None:
        cds = utils.fasta_to_dataframe('../genomes/human/Homo_sapiens.GRCh38.cds.all.fa')
    cds = cds.drop_duplicates('sequence')
    cds = cds[cds.sequence.str.len()>50]

    def split_seqs(r):
        #maxlen = int(np.random.normal(81,17))
        maxlen = int(np.random.gamma(9.5,9))
        #print len(r.sequence), maxlen
        s = [r.sequence[ind:ind+maxlen] for ind in range(0, len(r.sequence), maxlen)]
        return pd.Series(s)

    seqs = cds[:5000].apply(split_seqs,1).stack().reset_index(drop=True)
    seqs = seqs[seqs.str.len()>50]
    seqs = seqs[-seqs.str.contains('N')]
    result=[]
    i=1
    for seq in seqs:
        ms = int(np.random.randint(2,5))
        f = build_rna_features(seq, mature=seq[ms:ms+22])
        f['seq'] = seq
        if f['loops'] > 2:# or f['stem_length']>18:
            continue
        result.append(f)
        i+=1
        if i > samples:
            break
    result = pd.DataFrame(result)
    #result = result[(result.mfe*result.length<=-15)]
    result.to_csv('negative_mirna_features.csv', index=False)
    return result

def get_training_data(known=None, neg=None):
    """Get training data for classifier
        Args:
            known: known precursor data, a dataframe
            neg: negatives
        Returns:
            a dataframe with all features and a set of true/false values
    """

    if known is None:
        known = pd.read_csv(os.path.join(datadir, 'training_positives.csv'))
        #known = get_positives()
    if neg is None:
        neg = pd.read_csv(os.path.join(datadir, 'training_negatives.csv'))
    #print (len(known), len(neg))
    known['target'] = 1
    neg['target'] = 0
    data = pd.concat([known,neg]).reset_index(drop=True)
    data = data.sample(frac=1)
    y = data.target
    data = data.drop('target',1)
    X = data.select_dtypes(['float','int'])
    #X = sklearn.preprocessing.scale(X)
    return X, y

def precursor_classifier(known=None, neg=None, kind='regressor'):
    """Get a miRNA precursor classifier using given training data.
       Args:
        X: numpy array/dataframe with features
        y: true/false values matching feature rows, 1's and 0's
        kind: use 'classifier' or 'regressor' random forest
       Returns:
        random forest classifier fitted to X,y
    """

    X, y = get_training_data(known, neg)
    from sklearn.ensemble import (RandomForestClassifier, RandomForestRegressor)
    if kind == 'classifier':
        rf = RandomForestClassifier(n_estimators=100)
    else:
        rf = RandomForestRegressor(n_estimators=100)
    #print ('fitting..')
    rf.fit(X,y)
    return rf

def test_classifier(known=None, neg=None):

    from sklearn.ensemble import (RandomForestClassifier, RandomForestRegressor)
    X, y = get_training_data(known, neg)
    rf = RandomForestClassifier(n_estimators=50)
    rf.fit(X,y)
    from sklearn.model_selection import train_test_split,cross_val_score
    #X_train, X_test, y_train, y_test = train_test_split(X, y, train_size=0.4)
    #rf = RandomForestRegressor()
    scores = cross_val_score(rf, X, y, cv=5, scoring='roc_auc')
    print (scores)
    #print sklearn.metrics.classification_report(y_test, y_score)
    names = X.columns
    importances = rf.feature_importances_
    indices = np.argsort(importances)[::-1]
    print ('feature ranking:')
    for f in range(X.shape[1])[:10]:
        print("%d. %s (%f)" % (f + 1, names[indices[f]], importances[indices[f]]))
    '''a = neg[:2000].drop('target',1)
    a['score'] = score_features(a, rf)
    b = known[:2000].drop('target',1)
    b['score'] = score_features(b, rf)
    x = a.score.value_counts().sort_index()
    y = b.score.value_counts().sort_index()
    res = pd.DataFrame({'neg':x,'pos':y})
    res.plot(kind='bar')
    '''
    return

def score_features(data, rf):
    """Score a set of rna features"""

    X = data.select_dtypes(['float','int'])
    #data['score'] = rf.predict(X)
    return rf.predict(X)

def build_cluster_trees(reads, cluster_distance=2, min_size=2):
    """Build cluster tree of reads from a dataframe of locations e.g from
        a set of aligned reads from a sam file.
    Args:
        cluster_distance: Distance in basepairs for two reads to be in the same cluster;
       for instance 20 would group all reads with 20bp of each other
        min_size: Number of reads necessary for a group to be considered a cluster;
       2 returns all groups with 2 or more overlapping reads
    Returns:
        dict of ClusterTrees per chromosome
    """

    import collections
    from bx.intervals.cluster import ClusterTree
    cluster_trees = collections.defaultdict(lambda:
            ClusterTree(cluster_distance, min_size))
    for i, row in reads.iterrows():
        chrom = row['name']
        #print chrom, row.read_id, row.start, row.end
        cluster_trees[chrom].insert(row.start, row.end, row.name)
    return dict(cluster_trees)

def get_read_clusters(reads, cluster_distance=0, min_size=3, key='align_id'):
    """Assign reads to clusters from a dataframe from an alignment
      Args:
        reads: pandas dataframe of reads with start, end, read_id fields
        key: key to use for assigning reads to clusters, should be unique
      Returns:
        the dataframe with cluster numbers assigned
    """

    reads = reads.copy()
    if key in reads.columns:
        reads.set_index(key, inplace=True)
    #build clustertrees per chromosome of reads and store by index values
    clustertrees = build_cluster_trees(reads, cluster_distance, min_size)

    groups = []
    i=1
    for chrom, cltree in clustertrees.items():
        #print (chrom)
        for start, end, ids in cltree.getregions():
            #print (start, end, ids)
            c = reads.ix[ids].copy()
            c['cl_start'] = start
            c['cl_end'] = end
            c['cluster'] = i
            #remove wrongly added opposite strand reads
            c = c.groupby(['strand']).filter(lambda x: len(x) > 1)
            groups.append(c)
            i+=1
    df = pd.concat(groups)
    return df

def get_cluster_groups(rcl):
    """Get the clusters from reads with assigned cluster numbers"""

    clusts = rcl.groupby(['name','cluster','cl_start','cl_end','strand'])\
                            .agg({'reads':np.sum,'length':np.max})\
                            .reset_index()\
                            .rename(columns={'cl_start':'start','cl_end':'end'})
    print ('%s read clusters in %s reads' %(len(clusts),len(rcl)))
    clusts['clust_size'] = clusts.end-clusts.start
    return clusts

def generate_precursors(ref_fasta, coords, mature=None, step=5):
    """Create a set of possible precursor sequences from flanking sequence given
       genomic coordinates and the reference genome sequence.
       Args:
        ref_fasta: reference genome fasta file
        cooords: coordinates of estimated mature sequence to start from
        seq: mature sequence
       Returns:
        dataframe with precursor sequences and their coordinates
    """

    chrom,start,end,strand=coords
    loop = 15
    N = []
    if mature != None:
        seqlen = len(mature)
    else:
        seqlen = 22
    #generate candidate precursors
    for i in range(2,40,step):
        #5' side
        start5 = start - i
        end5 = start + 2 * seqlen-1 + loop + i
        coords = [chrom,start5,end5,strand]
        prseq = utils.sequence_from_coords(ref_fasta, coords)
        if prseq == None:
            continue
        struct,sc = utils.rnafold(prseq)
        #prseq, struct = check_hairpin(prseq, struct)
        mstatus = check_mature(prseq, struct, mature)
        #print (mstatus)
        #print (i)
        #print (prseq)
        #print (struct)
        N.append({'precursor':prseq,'struct':struct,'score':sc,
                  'chrom':chrom,'start':start5,'end':end5,
                  'mature':mature,'strand':strand, 'mature_check': mstatus})
    for i in range(2,40,step):
        #3' side
        start3 = start - (loop + seqlen + i)
        end3 = end + i + seqlen+1
        coords = [chrom,start3,end3,strand]
        prseq = utils.sequence_from_coords(ref_fasta, coords)
        if prseq == None:
            continue
        struct,sc = utils.rnafold(prseq)
        #prseq, struct = check_hairpin(prseq, struct)
        mstatus = check_mature(prseq, struct, mature)
        N.append({'precursor':prseq,'struct':struct,'score':sc,
                  'chrom':chrom,'start':start3,'end':end3,
                  'mature':mature,'strand':strand,'mature_check': mstatus})
    N = pd.DataFrame(N)
    return N

def score_precursors(N):
    """Filter/score a set of precursor sequences, requires a dataframe
      with a column of precursor sequences and structures made by the
      generate_precursors method. """

    global CLASSIFIER
    if CLASSIFIER == None:
        rf = precursor_classifier()
    else:
        rf = CLASSIFIER
    f = N.apply( lambda x: pd.Series(build_rna_features(x.precursor,
                                                        x.struct, x.score, x.mature)), 1 )
    #N['struct'] = f.struct
    #check mature out of stem range also
    #check mature for non templated additions?

    N['score'] = score_features(f, rf)
    N['mfe'] = f.mfe
    #filter by features
    N = N[(f.loops==1) & (f.stem_length>18) & (f.mfe*f.length<-15) & \
          (f.longest_bulge<10) & (f.gc<75) & (f.bulges_asymmetric<7)]

    N = N.sort_values('mfe')
    #print (N)
    return N

def get_consensus_read(ref, df):
    """Get consensus read from cluster of aligned reads"""

    strand = df.iloc[0].strand
    if strand == '+':
        s='start'
    else:
        s='end'
    g = df.groupby([s,'length']).agg({'reads':np.sum})\
        .sort_values(by='reads',ascending=False)\
        .reset_index()
    #print g
    x = g.iloc[0]
    cons = df[(df[s]==x[s]) & (df.length==x.length)].sort_values(by='reads',ascending=False)
    #print cons
    mature = None
    for i,r in cons.iterrows():
        if r.seq in ref:
            mature = r.seq
            break
    if mature==None:
        mature = cons.iloc[0].seq
    return mature

def find_precursor(ref_fasta, m, o=None, step=5, score_cutoff=.7):
    """Find the most likely precursor from a genomic sequence and
       one or two mapped read clusters.
       Args:
           ref_fasta: genomic reference sequence
           cluster: reads in a cluster, a dataframe
           cluster2: a pair to the first cluster, optional
           step: increment for extending precursors
           score_cutoff: if using non-classifier, optional
       Returns:
           the top precursor
    """

    x=m.iloc[0]
    rcoords = (x['name'], x.start-10, x.end+10, x.strand)
    refseq = utils.sequence_from_coords(ref_fasta, rcoords)
    #get a consensus mature sequence
    mature = get_consensus_read(refseq, m)

    coords = (x['name'], x.start, x.start, x.strand)
    N = generate_precursors(ref_fasta, coords, mature=mature, step=step)
    if len(N)==0:
        return
    N = score_precursors(N)
    #print ('%s candidates' %len(N))
    N = N[N.score>=score_cutoff]
    if len(N)>0:
        P = N.iloc[0].copy()
        #print(P)
    else:
        return

    #print (o)
    maturecounts = m.reads.sum()
    star = find_star_sequence(P.precursor, mature, P.struct)
    starcounts = 0
    if o is not None and star != None:
        #check reads from local cluster that are inside star seq?
        s = utils.find_subseq(P.precursor, star)
        #if P.strand == '+':
        ss = P.start+s; se = ss+len(star)
            #sreads = o[(o.start>=ss-2) & (o.end<=se+3)]
        #else:
            #se = P.end-s; ss = se-len(star)
        sreads = o[(o.start>=ss-2) & (o.end<=se+3)]
        starcounts = sreads.reads.sum()
        #print (ss, se)
        #print (sreads)
        #print display(HTML(forna_url(P.precursor, mature, star)))

    print (P.start, P.end, P.strand)
    print (mature, maturecounts, star, starcounts)
    P['mature_reads'] = maturecounts
    P['star_reads'] = starcounts
    P['star'] = star
    P['cluster'] = x.cluster
    #print (mature, star, maturecounts, starcounts)
    #print ('')
    return P

def find_mirnas(reads, ref_fasta, score_cutoff=.8, read_cutoff=50, species='',
                max_length=25, min_length=18, min_size=3):
    """Find novel miRNAs in reference mapped reads. Assumes we have already
        mapped to known miRNAs.
        Args:
            reads: unique aligned reads with counts in a dataframe
            ref_fasta: reference genome fasta file
            score_cutoff: max score to keep for precursors
            species: three letter mirbase code for species, optional
        Returns:
            tuple of dataframes: novel mirnas and the read clusters
    """

    global CLASSIFIER
    if CLASSIFIER == None:
        print ('getting default classifier')
        CLASSIFIER = precursor_classifier(kind='regressor')

    reads = reads[(reads.length<=max_length) & (reads.length>=min_length)]
    #assign reads to clusters
    rcl = get_read_clusters(reads, 10, min_size)
    clusts = get_cluster_groups(rcl)
    clusts = clusts[clusts.reads>=read_cutoff]
    print ('%s clusters above reads cutoff' %len(clusts))

    X = []
    N = []
    for i,c in clusts.iterrows():
        df = rcl[rcl.cluster==c.cluster]
        df = df.sort_values('reads',ascending=False)
        #small clusters should belong to a single mature
        if c.clust_size<28:
            df['mature'] = True
            X.append(df)
            p = find_precursor(ref_fasta, df, score_cutoff=score_cutoff)
            #print p
            if p is not None:
                N.append(p)
                #print (p.mature)
            continue

        #for larger clusters choose most likely mature reads
        anchor = df.iloc[0]
        st = anchor.start
        end = anchor.end
        m = df.loc[(abs(df.start-st)<=3) & (abs(df.end-end)<=5)].copy()
        if m.reads.sum() < read_cutoff:
            continue
        m['mature'] = True
        X.append(m)
        #remainder of reads assigned as non-mature
        o = df.loc[-df.index.isin(m.index)].copy()
        o['mature'] = False
        X.append(o)

        p = find_precursor(ref_fasta, m, o, score_cutoff=score_cutoff)

        #check for nearby clusters inside this precursor
        #nc = clusts[(abs(clusts.start-c.start)<100) & (clusts.cluster!=c.cluster)]

        if p is not None:
            N.append(p)
            #print (p.mature)

    new = pd.DataFrame(N)
    new['seed'] = new.apply(lambda x: x.mature[1:7], 1)
    #get coords column
    new['coords'] = new.apply(get_coords_string,1)
    new = new.reset_index(drop=True)
    assign_names(new, species)
    kp = base.get_mirbase(species)
    #check if the mature are already in known precursors
    new['known_id'] = new.apply( lambda x: find_from_known_prec(x, kp), 1)
    new = new.sort_values(by='mature_reads', ascending=False)

    print ('found %s novel mirnas' %len(new.groupby('mature_id')))
    print ('%s with known mature sequences' %len(new[-new.known_id.isnull()]))
    #also return all the reads found in clusters
    found = pd.concat(X)
    return new, found

def get_coords_string(r):
    """coords string from fields"""
    if 'chrom' not in r:
        r['chrom']=r['name']
    return r['chrom']+':'+str(r.start)+'..'+str(r.end)+':'+r.strand

def find_from_known_prec(x, prec):
    for i,r in prec.iterrows():
        if r.precursor.find(x.mature[:-1]) != -1:
            return r.mirbase_id

def assign_names(df, species=''):
    """Assign name to novel mirna, precursor/mature ids should allow consistent
       identification across datasets"""

    df['precursor_id'] = df.apply( lambda x: species+'_novel_'+x.chrom+'_'+str(x.start),1 )
    df['mature_id'] = df.apply( lambda x: species+'_'+encode_name(x.mature), 1 )
    return

def encode_name(s):
    """hash a sequence into a short string"""

    import hashlib
    h = hashlib.md5(s.encode())
    s = h.digest().encode('base64')[:8]
    s = s.replace('/','x')
    return s

def forna_url(precursor, mature, star=None, struct=None):
    """Create link to view mirna structure in forna web app"""

    #print x
    pd.options.display.max_colwidth = 500
    #seq = x.precursor
    #mature = x[col1]
    mstart = utils.find_subseq(precursor, mature)+1
    mend = mstart+len(mature)-1
    if struct==None:
        struct = utils.rnafold(precursor)[0]
    colors = '%s-%s:lightgreen\\n' %(mstart, mend)
    if star != None:
        #print x[maturecol], x[col2]
        #star = x[col2]
        if star != None and not pd.isnull(star):
            sstart = utils.find_subseq(precursor, star)+1
            send = sstart+len(star)-1
            colors += '%s-%s:pink' %(sstart,send)
    url='http://nibiru.tbi.univie.ac.at/forna/forna.html'\
        '?id=url/name&sequence=%s&structure=%s&colors=%s' %(precursor,struct,colors)
    html = '<a href="%s" target="_blank"> view structure </a>' %url
    return html

def string_to_html(s):
    """Convert lines of strings for html rendering"""

    html=''
    x = s.split('\n')
    for line in x:
        line = line.replace(' ','&nbsp') #preserve spaces
        html += line+'<br>'
    return html

def create_report(df, reads, species=None, outfile='report.html'):
    """Novel miRNA predictions html report"""

    pd.options.display.max_colwidth = 500
    css = get_css()
    h = '<html><head><meta charset="utf-8">  <title>novel miRNA</title>'
    h += '<style media="screen" type="text/css"> %s </style>' %css
    h += '</head>'
    h += '<body>'
    h += '<div class="header">'
    h += '<h3>novel miRNA predictions</h3>'
    h += '</div>'
    h += '<div class="sidebar">'
    links = df[['mature_id','mature_reads','chrom','start']].copy()
    links['mature_id'] = links.mature_id.apply(lambda x: ('<a href=#%s > %s </a>' %(x,x)))
    links = links.set_index(['mature_id','chrom','start'])
    h += links.to_html(escape=False, classes='sidebar', sparsify=True)#, index=False)
    h += '</div>'

    df = df.copy()
    df = df.set_index('mature_id')

    ens_sp = pd.read_csv(os.path.join(datadir, 'ensembl_names.csv'), index_col=0)
    if species in ens_sp.index:
        ensname = ens_sp.ix[species]['scientific name']
        df['coords'] = df.coords.apply(
            lambda x: ('<a href=http://www.ensembl.org/%s/Location/View?r=%s target="_blank">%s </a>' %(ensname,x,x)),1)
    df['link'] = df.apply(lambda x: forna_url(x.precursor, x.mature, x.star, x.struct), 1)

    h += '<div class="content">'
    for i,r in df.iterrows():
        #print (r.name, r.mature, r.star)
        h += '<div class="box">'
        h += '<a name=%s></a>' %i
        h += r.to_frame().to_html(escape=False)
        x = reads[(reads.cluster==r.cluster)]
        #print (x)
        #print (x.reads.sum())
        s = utils.print_read_stack(x, r.precursor, by='reads')
        if s==None:
            h+='</div>'
            continue
        h += '<p>'
        h += string_to_html(s)
        h += '</p>'
        h += '</div>'

    h+='</div>'
    h += '</body>'
    f=open(outfile,'w')
    f.write(h)
    f.close()
    return h

def get_css():
    """Get css style for embedding in html page"""

    fname = os.path.join(datadir, 'styles.css')
    with open(fname) as f:
        content = f.readlines()
        content = ''.join(content)
    return content



from setuptools import setup
import sys,os

inst_requires = ['pandas>=0.17',
                    'seaborn>=0.7',
                    'scikit-learn>=0.18',
                    'pybedtools>=0.7.9',
                    'pysam>=0.10.0',
                    'HTSeq>=0.6',
                    'bx-python>=0.5',
                    'forgi>=0.4']

major, minor, micro = sys.version_info[:3]
if major == '2':
    inst_requires.append('future')

setup(
    name = 'smallrnaseq',
    version = '0.4.0',
    description = 'Package for short RNA-seq analysis',
    long_description = 'smallrnaseq is a Python package for processing of small RNA seq data.',
    url='https://github.com/dmnfarrell/smallrnaseq',
    license='GPL v3',
    author = 'Damien Farrell',
    author_email = 'farrell.damien@gmail.com',
    packages = ['smallrnaseq'],
    package_data={'smallrnaseq': ['data/*','*.R']},
    install_requires=inst_requires,
    entry_points = {
        'console_scripts': [
            'smallrnaseq=smallrnaseq.app:main',
            'mirdeep2=smallrnaseq.mirdeep2:main']
            },
    classifiers = ['Operating System :: OS Independent',
            'Programming Language :: Python :: 3.4',
            'Programming Language :: Python :: 2.7',
            'Operating System :: MacOS :: MacOS X',
            'Operating System :: Microsoft :: Windows',
            'Operating System :: POSIX',
            'License :: OSI Approved :: Apache Software License',
            'Development Status :: 4 - Beta',
            'Intended Audience :: Science/Research'],
    keywords = ['rna','sequencing','mirdeep2','biology','scientific'],
)



#!/usr/bin/env python

"""
    Module for core smallrnaseq functions
    Created July 2014
    Copyright (C) Damien Farrell

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 3
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
"""

from __future__ import absolute_import, print_function
import sys, os, string, time
import types, re, subprocess, glob, shutil
import numpy as np
import pandas as pd
try:
    import HTSeq
except:
    'HTSeq not present'
from . import utils, aligners

path = os.path.dirname(os.path.abspath(__file__)) #path to module
datadir = os.path.join(path, 'data')
MIRBASE = os.path.join(datadir, 'miRBase_all.csv')


def first(x):
    return x.iloc[0]

def get_exons(gtf):
    """Get exon featues into GenomicArrayOfSets for HTSeq
       counting"""

    exons = HTSeq.GenomicArrayOfSets( "auto", stranded=True )
    for feature in gtf:
        if feature.type == "exon":
            if 'transcript_id' in feature.attr:
                exons[ feature.iv ] += feature.attr["transcript_id"]
            else:
                pass
    return exons

def gtf_to_dataframe(gtf=None, gtf_file=None, index='transcript_id'):
    """Convert gtf/gff features to a pandas dataframe"""

    if gtf_file != None:
        gtf = HTSeq.GFF_Reader(gtf_file)
    recs=[]
    for f in gtf:
        r = {'id':f.name, 'chrom':f.iv.chrom, 'start':f.iv.start,
             'end': f.iv.end, 'strand':f.iv.strand}
        r.update(f.attr)
        recs.append(r)
    df = pd.DataFrame(recs)
    try:
        df = df.drop_duplicates(['exon_id'])
    except:
        print ('no exon_id field')
    return df

def count_features(samfile, features=None, gtffile=None, readcounts=None, merge=False):
    """Count reads in features from an alignment, if no truecounts we
       assume a non-collapsed file was used to map
       Args:
           samfile: mapped sam file
           gtffile: feature file
           features: annotations read from bed or gtf file
           truecounts: read counts from original (un-collapsed) file
           merge: whether to merge the gtf fields with the results
       Returns: dataframe of genes with total counts
    """

    if gtffile != None:
        gtf = HTSeq.GFF_Reader(gtffile)
        features = get_exons(gtf)
    sam = HTSeq.SAM_Reader(samfile)
    if type(readcounts) is pd.DataFrame:
        readcounts = {r.seq: r['reads'] for i,r in readcounts.iterrows()}
    import collections
    counts = collections.Counter()
    for almnt in sam:
        seq = str(almnt.read)
        if readcounts is not None and seq in readcounts:
            c = readcounts[seq]
        else:
            c = 1
        if not almnt.aligned:
            counts[ "_unmapped" ] += c
            continue
        gene_ids = set()
        for iv, val in features[ almnt.iv ].steps():
            gene_ids |= val
        #print almnt.iv, almnt.read, readcounts[seq], gene_ids
        if len(gene_ids) == 1:
            gene_id = list(gene_ids)[0]
            counts[ gene_id ] += c
        elif len(gene_ids) == 0:
            counts[ "_no_feature" ] += c
        else:
            counts[ "_ambiguous" ] += c
    result = pd.DataFrame(counts.items(),columns=['name','reads'])
    #result['norm'] = result.reads/result.reads.sum()*1e6
    result = result.sort_values(by='reads',ascending=False)

    um = ['_no_feature','_unmapped']
    mapped = float(result[-result.name.isin(um)].reads.sum())
    total = result.reads.sum()

    print ('%s/%s reads counted, %.2f percent' %(mapped, total, mapped/total*100))
    if merge == True and gtffile != None:
        result = merge_features(result, gtffile)
    return result

def merge_features(counts, gtffile):
    """Merge counts with dataframe containing original gtf info"""

    gtf = HTSeq.GFF_Reader(gtffile)
    df = gtf_to_dataframe(gtf)
    hits = counts.merge(df, left_on='name', right_on='transcript_id', how='inner')
    #hits = hits.sort_values(by='reads',ascending=False)
    return hits

def feature_counts_summary(counts):
    """Summary of feature counts by gene biotype"""

    counts['perc'] = counts.norm/1e6*100
    s = counts.groupby('gene_biotype').agg({'reads':np.sum,'perc':np.sum})
    ax = s.plot(y='perc',kind='barh',figsize=(5,3))
    return s

def get_top_genes(counts):

    df = counts.groupby('gene_name')\
                .agg({'reads':sum,'transcript_id':np.size})\
                .sort_values('reads',ascending=False)
    return df

def count_aligned(samfile, collapsed=None, readcounts=None, by='name',
                   count_method='split'):
    """Count short read alignments from a sam or bam file. Designed to be used with
       collapsed reads with original read counts in fasta id.
       Args:
           samfile: mapped sam file
           collapsed: collapsed fasta with original read counts in name
           readcounts: dataframe with original read counts, optional
           by: whether to group the counts by 'name' (default) or 'seq'
           count_method: method to handle multi mapping reads, default is 'split'
            which divides counts equally over each alignment
    """

    if collapsed != None:
        readcounts = utils.read_collapsed_file(collapsed)

    counts = utils.get_aligned_reads(samfile, readcounts=readcounts)
    if count_method == 'split':
        #split multi mapped sequence counts evenly
        counts['reads'] = counts.groupby(['seq'])['reads'].transform(lambda x: x / len(x))

    #if drop_duplicates == True:
    #    counts = counts.drop_duplicates('seq')
    total = readcounts.reads.sum()
    counts = counts.groupby('name').agg({'reads':np.sum}).reset_index()
    mapped = float(counts.reads.sum())
    if len(counts) > 0:
        print ('%s/%s reads counted, %.2f percent' %(mapped, total, mapped/total*100))
    else:
        print ('no counts found')
    return counts

def pivot_count_data(counts, idxcols='name', norm_method='library', sortby=None):
    """Pivot read counts created by count_aligned over multiple samples
       and get normalised read counts.
       Args:
           counts: dataframe of raw count data with samples per column
           idxcols: name of index column
           norm_method: how to normalize the counts (returned in extra columns),
                        default is total library counts
       Returns: dataframe of raw /normalised read counts with column per sample
    """

    x = pd.pivot_table(counts, values='reads', index=idxcols, columns='label')
    x = x.round(0)
    #x = x.fillna(0).astype(int)
    if norm_method == 'library':
        n = total_library_normalize(x)
    elif norm_method == 'quantile':
        n = quantile_normalize(x)

    #x.columns = [i+'_' for i in x.columns]
    scols = x.columns
    ncols = n.columns = [i+' norm' for i in n.columns]
    x = x.join(n)
    #scols,ncols = get_column_names(x)
    x['total_reads'] = x[scols].sum(1)
    x['mean_norm'] = x[ncols].apply(lambda r: r[r.nonzero()[0]].mean(),1)
    x = x.reset_index()
    #print (x[50:])
    if sortby != None:
        x = x.sort_values(by=sortby, ascending=False)
    return x

def normalize_samples(counts, norm_method='library', rename=True):
    """Normalize over a matrix of samples explicitly, this will overwrite any 'norm'
       columns created previously when pivoting the count data
       Args:
            counts: dataframe of raw count data with samples per column
            rename: rename columns with 'norm' label and add to existing ones
       Returns: dataframe of raw /normalised read counts
    """

    x = counts
    if norm_method == 'library':
        n = total_library_normalize(x)
    elif norm_method == 'quantile':
        n = quantile_normalize(x)
    if rename == True:
        scols = x.columns
        ncols = n.columns = [i+' norm' for i in n.columns]
        x = x.join(n)
    else:
        x = n
    return x

def get_column_names(df):
    """Get count data sample column names"""

    ignore = ['total_reads','mean_norm']
    ncols = [i for i in df.columns if (i.endswith('norm')) and i not in ignore]
    cols = [i.split(' ')[0] for i in ncols if i not in ignore]
    return cols, ncols

def total_library_normalize(df):
    """Normalise by size of total reads"""

    df = df.copy()
    for col in df:
        df[col] = df[col]/df[col].sum()*1e6
    df = df.round(2)
    return df

def quantile_normalize(df):
    """Quantile normlization of counts for multiple samples.
       see https://github.com/ShawnLYU/Quantile_Normalize
    """

    df = df.copy()
    #compute rank
    dic = {}
    for col in df:
        dic.update({col : sorted(df[col])})
    sorted_df = pd.DataFrame(dic)
    rank = sorted_df.mean(axis = 1).tolist()
    #sort
    for col in df:
        t = np.searchsorted(np.sort(df[col]), df[col])
        df[col] = [rank[i] for i in t]
    df=df.round(2)
    return df

def upper_quartile_normalize(df):
    """Upper quartile noralisation"""

    u = x.apply(lambda r: r[r>r.quantile(0.75)])
    df = df/u.mean()
    return df*1e6

def deseq_normalize(df):
    """Compute count/geometric mean per sample"""

    df = df.copy()
    df = df.apply(lambda r: r/r.mean(),1)
    return df

def map_rnas(files, indexes, outpath, collapse=True, adapters=None, aligner='bowtie',
             norm_method='library', use_remaining=True, overwrite=False,
             samplelabels=None, params={}, count_method='split'):
    """Map reads to one or more gene annotations, assumes adapters are removed
    Args:
        files: input fastq read files
        indexes: bowtie indexes of annotations/genomes
        adapters: if adapters need to be trimmed
        use_remaining: only align to remaining reads after each index
        overwrite: whether to overwrite temp files
        samplelabels: mapping of file names to short ids
        params: index specific aligner parameters
    """

    if not os.path.exists(aligners.BOWTIE_INDEXES):
        print ('index folder does not exist!')
        return
    if not os.path.exists(outpath):
        os.mkdir(outpath)
    if overwrite == True:
        print ('removing old temp files')
        utils.remove_files(outpath,'*_mapped.sam')
        utils.remove_files(outpath, '*_r.fa')
    if collapse == True:
        cfiles = collapse_files(files, outpath)
    else:
        cfiles = files
    if len(cfiles)==0:
        print ('WARNING no files to align')
        return
    print ('')
    result = []
    #store current aligner parameters as default
    default_params = aligners.get_current_params(aligner)

    for cfile in cfiles:
        rem = None
        filename = os.path.splitext(os.path.basename(cfile))[0]
        print (cfile)
        readcounts = utils.read_collapsed_file(str(cfile))
        total = readcounts.reads.sum()
        print (filename)
        for idx in indexes:
            print (idx)
            if idx in params:
                align_params = params[idx]
            else:
                align_params = default_params
            aligners.set_params(aligner, align_params)
            #print (align_params)
            if use_remaining == True and rem != None:
                query = rem
            else:
                query = cfile
            samfile = str(os.path.join(outpath, '%s_%s.sam' %(filename,idx)))
            rem = os.path.join(outpath, filename+'_r.fa')

            if aligner == 'bowtie':
                aligners.bowtie_align(query, idx, outfile=samfile,
                                      remaining=rem, verbose=True)
            elif aligner == 'subread':
                aligners.subread_align(query, idx, samfile)

            counts = count_aligned(samfile, readcounts=readcounts, count_method=count_method)
            if len(counts) == 0:
                print ('WARNING: no counts found for %s.' %idx)
                continue
            if samplelabels != None:
                counts['label'] = samplelabels[filename]
            else:
                counts['label'] = filename
            counts['ref'] = idx
            counts['fraction'] = counts.reads/total
            #print (counts.reads.sum(), total)
            result.append(counts)
            print()
        print()
    if len(result) == 0:
        return
    result = pd.concat(result)
    counts = pivot_count_data(result, idxcols=['name','ref'], norm_method=norm_method,
                              sortby='total_reads')
    print ('done')
    return result, counts

def map_genome_features(files, ref, gtf_file, outpath='', aligner='bowtie',
                        overwrite=True, aligner_params=''):
    """Map multiple files to a genome with features and return/process hits.
       Can be used for miRNA discovery
       Args:
           ref: genome bowtie index name
           gtf_file: gtf or bed file with features
           bowtie _index: path with bowtie indexes
    """

    if aligner_params != '' :
        aligners.set_params(aligner, aligner_params)
    if overwrite == True:
        print ('removing old temp files')
        utils.remove_files(outpath,'*_mapped.sam')
        utils.remove_files(outpath, '*_r.fa')

    ext = os.path.splitext(gtf_file)[1]
    if ext == '.gtf' or ext == '.gff' or ext == '.gz':
        features = HTSeq.GFF_Reader(gtf_file)
    elif ext == '.bed':
        features = HTSeq.BED_Reader(gtf_file)

    exons = get_exons(features)

    cfiles = collapse_files(files, outpath)
    print (cfiles)
    result = []
    for cfile in cfiles:
        label = os.path.splitext(os.path.basename(cfile))[0]
        samfile = os.path.join(outpath, '%s_%s.sam' %(label,ref))
        if aligner == 'bowtie':
            aligners.bowtie_align(cfile, ref, outfile=samfile)
        elif aligner == 'subread':
            aligners.subread_align(cfile, ref, samfile)
        #get true read counts for collapsed file
        readcounts = utils.read_collapsed_file(cfile)
        #count features
        counts = count_features(samfile, features=exons, readcounts=readcounts)
        counts['label'] = label
        counts['genome'] = ref
        total = readcounts.reads.sum()
        counts['fraction'] = counts.reads/total
        result.append(counts)
    result = pd.concat(result)
    result = merge_features(result, gtf_file)
    return result

def get_base_names(files):
    names = [os.path.splitext(os.path.basename(f))[0] for f in files]
    return names

def assign_sample_ids(names, outfile='sample_labels.csv'):
    """Assign new ids for sample filenames e.g. files. Useful for
       replacing long file names with short ids.
       Returns: dict of filename/id values
    """

    i=1
    labels = {}
    for fname in names:
        n = os.path.splitext(os.path.basename(fname))[0]
        print (n)
        sid = 's%02d' %i
        labels[n] = sid
        i+=1
    l = pd.DataFrame.from_dict(labels,orient='index')
    l.columns = ['id']; l.index.name='filename'
    l.to_csv(outfile)
    return labels

def trim_files(files, outpath, adapters):
    """Trim adapters from fastq files"""

    for f in files:
        cut = os.path.join(outpath, f)
        if not os.path.exists(cut):
            trim_adapters(f, adapters, cut)
    return

def collapse_reads(infile, outfile=None, min_length=15, progress=False):
    """Collapse identical reads, retaining copy number in a csv file
       and writing collapsed reads to a new fasta file.
       Creates a collapsed fasta file of unique reads and a csv
       file with the """

    #from itertools import islice
    if outfile == None:
        outfile = os.path.splitext(infile)[0]+'_collapsed.fa'
    print ('collapsing reads %s' %infile)
    ext = os.path.splitext(infile)[1]
    if ext == '.fastq':
        fastfile = HTSeq.FastqReader(infile, "solexa")
    elif ext == '.fa' or ext == '.fasta':
        fastfile = HTSeq.FastaReader(infile)
    else:
        print ('not fasta or fastq')
        return False

    i=0
    total = 0
    f = {}
    #print (fastfile)
    for s in fastfile:
        seq = s.seq.decode()
        if seq in f:
            f[seq]['reads'] += 1
        else:
            f[seq] = {'name':s.name, 'reads':1}
        total += 1

    df = pd.DataFrame.from_dict(f, orient='index')
    df.index.name = 'seq'
    df = df.reset_index()
    l = df.seq.str.len()
    df = df[l>=min_length]
    df = df.drop(['name'], 1)
    df = df.sort_values(by='reads', ascending=False).reset_index()
    df['read_id'] = df.index.copy()
    df['read_id'] = df.apply( lambda x: str(x.read_id)+'_'+str(x.reads), 1 )
    #print df[:10]
    utils.dataframe_to_fasta(df, idkey='read_id', outfile=outfile)
    #df.to_csv(os.path.splitext(outfile)[0]+'.csv', index=False)
    print ('collapsed %s reads to %s' %(total,len(df)))
    return True

def collapse_files(files, outpath, **kwargs):
    """Collapse reads and save counts
        min_length: min length of reads to include
    """

    outfiles = []
    for f in files:
        label = os.path.splitext(os.path.basename(f))[0]
        collapsedfile = os.path.join(outpath, label+'.fa')
        if not os.path.exists(collapsedfile):
            res = collapse_reads(f, outfile=collapsedfile, **kwargs)
            if res == False:
                continue
        else:
            print ('found collapsed file %s' %collapsedfile)
        outfiles.append(collapsedfile)
    return outfiles

def _get_mature(r, idkey='mature1', seqkey='mature1_seq', pad5=0, pad3=0):
    """Get mature sequences from mirbase file,
      this is a row-based dataframe function"""

    p = r.precursor
    name = r[idkey]
    m = r[seqkey]
    if pd.isnull(m):
        s = np.nan
    else:
        i = p.find(m)
        start = i-pad5
        if start < 0: start = 0
        s = p[start:i+len(m)+pad3]
    return pd.Series([name,s],index=['name','sequence'])

def get_mature_padded(df, **kwargs):

    m = df.apply(lambda x: _get_mature(x,**kwargs), 1)
    return m

def get_mirbase(species=None):
    """Get mirbase sequences"""

    df = pd.read_csv(MIRBASE)
    if species is not None:
        df = df[df.species==species]
    df.precursor = df.precursor.str.replace('U','T')
    df.mature1_seq = df.mature1_seq.str.replace('U','T')
    df.mature2_seq = df.mature2_seq.str.replace('U','T')
    return df

def get_mirbase_sequences(species='hsa', pad5=0, pad3=0, dna=False):
    """Extract species specific sequences from mirbase file.
       Args:
           species: 3-letter code for species
           n: bases to extend around ends of mature sequence
       Returns:
            dataframe with mature or hairpin sequences
    """

    df = pd.read_csv(MIRBASE)
    if species != None:
        df = df[df.species==species]

    #get both 5p and 3p seqs for each mirna
    m1 = df.apply(lambda x: _get_mature(x, 'mature1', 'mature1_seq', pad5, pad3), 1)
    m2 = df.apply(lambda x: _get_mature(x, 'mature2', 'mature2_seq', pad5, pad3), 1)
    df = pd.concat([m1,m2]).dropna().reset_index(drop=True)
    df = df[df.sequence.str.len()>2]
    df = df.drop_duplicates('name')
    if dna == True:
        df['sequence'] = df.sequence.str.replace('U','T')
    return df

def build_mirbase_index(species, aligner='bowtie', pad5=3, pad3=5,
                        kind='mature', index_path='indexes'):
    """Build species-specific mirbase bowtie index
       Args:
           species: 3-letter code for species
           aligner: which aligner to build for
           n: bases to extend around ends of mature sequence
    """

    if kind == 'mature':
        mirs = get_mirbase_sequences(species, pad5, pad3)
        idxname = 'mirbase-'+species
        outfile = '%s.fa' %idxname
        utils.dataframe_to_fasta(mirs, seqkey='sequence', idkey='name',
                                outfile=outfile)
    else:
        mirs = get_mirbase(species)
        idxname = 'precursors-'+species
        outfile = '%s.fa' %idxname
        utils.dataframe_to_fasta(mirs, seqkey='precursor', idkey='mirbase_id',
                                outfile=outfile)
        print ('got %s sequences' %len(mirs))

    if aligner == 'bowtie':
        aligners.build_bowtie_index(outfile, index_path)
    elif aligner == 'subread':
        aligners.build_subread_index(outfile, index_path)
    return idxname

def map_mirbase(files, species='bta', outpath='mirna_results', indexes=[],
                ref_genome='',
                overwrite=False, aligner='bowtie',
                pad5=3, pad3=5, **kwargs):
    """Map multiple fastq files to mirbase mature sequences and get
       count results into one file. Used for counting of known miRNAs.
       Args:
           Species: three letter name of species using mirbase convention
           indexes: other libraries to align to before we count mirnas
    """

    if not os.path.exists(outpath):
        os.mkdir(outpath)
    index_path = aligners.BOWTIE_INDEXES
    if index_path == None:
        index_path = 'indexes'
    #generate new mirbase bowtie index
    if aligner == 'bowtie':
        if aligners.BOWTIE_PARAMS == None:
            aligners.BOWTIE_PARAMS = '-n 1 -l 20'
    elif aligner == 'subread':
        #aligners.SUBREAD_INDEXES = index_path
        aligners.SUBREAD_PARAMS = '-m 2 -M 2'

    midx = build_mirbase_index(species, aligner, pad5, pad3, index_path=index_path)
    pidx = build_mirbase_index(species, aligner, kind='precursor', index_path=index_path)

    indexes.extend([midx, pidx])
    #add the reference genome to be aligned last if provided
    if ref_genome != '':
        indexes.append(ref_genome)
    #now map to the mirbase index for all files
    res, counts = map_rnas(files, indexes, outpath, overwrite=overwrite,
                           aligner=aligner, **kwargs)
    return res, counts

def map_isomirs(files, outpath, species, samplelabels=None):
    """Count mirna isomirs using previously aligned files"""

    idx = 'mirbase-'+species
    result = []
    for f in files:
        filename = os.path.splitext(os.path.basename(f))[0]
        samfile = os.path.join(outpath, '%s_%s.sam' %(filename,idx))
        #countsfile = os.path.join(outpath, '%s.csv' %filename)
        collapsed = os.path.join(outpath, '%s.fa' %filename)
        c = count_isomirs(samfile, collapsed, species)
        if samplelabels != None:
            c['label'] = samplelabels[filename]
        else:
            c['label'] = filename
        result.append(c)
    result = pd.concat(result)
    counts = pivot_count_data(result, idxcols=['name','seq'], sortby='total_reads')
    return result, counts

def _get_iso_class(x, refs, crefs):
    """Get isomir class using sRNAbench type scheme"""

    name = x['name']
    ref = refs.ix[name]
    cseq = crefs.ix[name].sequence
    cstart = utils.find_subseq(ref.sequence, cseq)
    cend = cstart+len(cseq)
    start = utils.find_subseq(ref.sequence, x.seq)
    end = start+len(x.seq)
    xs = int(start-cstart)
    ys = int(end-cend)
    nuctempl = ref.sequence[end-ys:end]
    nucend = x.seq[-ys:]
    subclass=''
    if start == cstart:
        if end == cend:
            isoclass = 'exact'
            if x.seq == cseq:
                subclass = 'exact'
            else:
                subclass = 'exactnucvar'
        elif end>cend and nucend != nuctempl:
            isoclass = 'nta'
            subclass = 'nta#%s' %nucend
        else:
            isoclass = 'lv'
            subclass = 'lv3p|%s' %ys
    elif end == cend:
        isoclass = 'lv'
        subclass = 'lv5p|%s' %xs
    else:
        #print start, cstart
        #print x.seq, cseq, ref.sequence, name
        isoclass = 'mv'
        subclass = 'lv5p|%s|lv3p|%s' %(xs,ys)

    return pd.Series([isoclass,subclass],index=['isoclass','subclass'])

def count_isomirs(samfile, collapsed, species):
    """Count miRNA isomirs using aligned reads from a samfile and actual
       read counts from a csv file"""

    #get canonical mirna mature sequences from mirbase
    canonical = get_mirbase_sequences(species, dna=True).set_index('name')
    #padded sequences so we can see where each read landed relative to canonical
    mirs = get_mirbase_sequences(species, pad5=6, pad3=6, dna=True).set_index('name')
    reads = utils.get_aligned_reads(samfile, collapsed)
    reads = reads.drop(['read_id','start','end'],1)
    x = reads.apply(lambda x: _get_iso_class(x, mirs, canonical),1)
    reads = reads.join(x)
    reads['name'] = reads.name + '_' + reads.subclass
    return reads

def output_read_stacks(files, ref_fasta, outfile='read_stacks.txt'):
    """Output read stack files from sam alignment files and a ref sequence"""

    out = open(outfile, 'w')
    s=''
    for f in files:
        filename = os.path.splitext(os.path.basename(f))[0]
        samfile = os.path.join(outpath, '%s_%s.sam' %(filename,idx))
        countfile = os.path.join(outpath, '%s.csv' %filename)
        readcounts = pd.read_csv(countfile)
        reads = utils.get_aligned_reads(samfile, readcounts)
        s += filename+'\n'
        s += utils.print_read_stacks(reads, fastafile='mirbase-bta.fa')
    #print (s)
    out.write(s)
    out.close()
    return

def filter_expr_results(df, freq=0.5, meanreads=0, totalreads=50):

    c,normcols = getColumnNames(df)
    df = df[df.freq>freq]
    df = df[df['total']>=totalreads]
    df = df[df['mean_norm']>=meanreads]
    return df

def compare_expression_profiles(df, by='ref', key='reads', threshold=1, path=None):
    """Scatter matrix of count values across samples and/or by label"""

    df = df[df.name!='_unmapped']
    df = df[df[key]>=threshold]
    X = pd.pivot_table(df, values=key, index=['name',by], columns=['label'])
    X = np.log(X).dropna()
    X = X.reset_index(1)
    g = sns.pairplot(X,hue=by)
    if path != None:
        plt.savefig(os.path.join(path, 'expr_scatter.png'))
    return

def get_fractions_mapped(df, by=['ref','label']):
    """Process results of multiple mappings to get fractions
    of each annotations mapped"""

    x = df.groupby(by).agg({'fraction':np.sum})
    x = x.unstack(level=0)
    x.columns = x.columns.droplevel(0)
    x['unmapped'] = 1-x.sum(1)
    #x = x.T
    #x = x.reindex_axis((x).mean(1).sort_values().index)
    x = x.reset_index()
    return x



#!/usr/bin/env python

"""
    Module for methods using ensembl mostly using pycogent API
    Created September 2014
    Copyright (C) Damien Farrell

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 3
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
"""

from __future__ import absolute_import, print_function
import numpy as np
import pandas as pd
from . import base

try:
    import cogent
    from cogent.db.ensembl import HostAccount, Genome, Compara, Species
    account = None
    release = '87'
    #add recent ensembl species
    species = ['cow','human','mouse','rat','chimp','gorilla','orangutan',
               'macaque','dog','pig','cat','olive baboon','sheep']
    Species.amendSpecies('Papio anubis', 'olive baboon')
    Species.amendSpecies('Ovis aries', 'sheep')
    Species.amendSpecies('Erinaceus europaeus', 'hedgehog')
    Species.amendSpecies('Mustela putorius furo', 'ferret')
except Exception as e:
    print (e)


def get_orthologs(refgenome, ensid=None, symbol=None):
    '''if ensid!=None:
        mygene = cow.getGeneByStableId(StableId=ensid)
    else:
        mygene = cow.getGenesMatching(symbol=symbol)'''
    #print mygene
    orthologs = comp.getRelatedGenes(gene_region=mygene,
                    Relationship='ortholog_one2one')
    print (orthologs.Members)
    #seqs = orthologs.getSeqCollection(feature_types='gene')
    #print seqs.Names
    #syntenicregions = comp.getSyntenicRegions(region=mygene,
    #                align_method='EPO', align_clade='eutherian')
    return

def get_genes_from_location(ref, coords, pad=0):
    """Get genes from a set of genome coordinates.
       pad will add n bases to either side to expand area"""

    genome = Genome(Species=ref, Release=release, account=account)
    chrom,start,end,strand = coords
    genes = list(genome.getFeatures(CoordName=chrom, Start=start-pad,
                    End=end+pad, feature_types='gene'))
    return genes

def find_in_gene(g, start, end):
    """Find if coords are inside intron or exon"""

    start=int(start)
    end=int(end)
    tr = g.CanonicalTranscript
    for i in tr.Exons:
        s,e = i.Location.Start,i.Location.End
        if start-5>=s and end<=e+5:
            return 'exon'
    if tr.Introns is None: return
    for i in tr.Introns:
        s,e = i.Location.Start,i.Location.End
        if start-5>=s and end<=e+5:
            return 'intron'
    #return 'both'

def get_alignment_tree(fname):
    """Build a neighbour joining tree"""

    from cogent.phylo import distance, nj
    from cogent.evolve.models import HKY85,F81
    al = cogent.LoadSeqs(fname,format='fasta')
    d = distance.EstimateDistances(al, submodel= F81())
    d.run(show_progress=False)
    mytree = nj.nj(d.getPairwiseDistances())
    mytree = mytree.balanced()
    print (mytree.asciiArt())
    print
    '''from cogent.draw import dendrogram
    p = dendrogram.SquareDendrogram(mytree)
    p.drawToPDF('tree-scaled.pdf', 500, 400, stroke_width=2.0,
                shade_param = 'r', max_value = 1.0,)'''
    return

def get_syntenic_alignment(comp, ref, coords, fname='ensembl.aln.fa'):
    """Get ensembl genomic alignment for a location using compara and return seqs"""

    print
    chrom,start,end,strand = coords
    regions = comp.getSyntenicRegions(Species=ref, CoordName=chrom,
                                       Start=start, End=end, align_method="EPO_LOW_COVERAGE",
                                       align_clade="39 eutherian", Strand=strand)
    regions = [r for r in regions]
    if len(regions)==0:
        print ('no alignments')
        return None, None
    #usually only 1 alignment
    A = regions[0].getAlignment()
    A.writeToFile(fname)
    #print cogent.LoadSeqs(fname)
    #if len(A.Seqs)>3:
    #    getAlignmentTree(fname)
    if regions is not None:
        print ('%s syntenic regions found' %len(regions))
    return regions, A

def get_host_genes(df, ref='cow'):
    """Get all genes containing the given miRNAs using ensembl"""

    results=[]
    comp = Compara(species, account=None, Release='79')
    for i, r in list(df.iterrows()):
        name = r['#miRNA']
        c,locs,strand = r['precursor coordinate'].split(':')
        start,end = locs.split('..')
        coords = c,int(start),int(end),strand
        genes = get_genes_from_location(ref, coords)
        for g in genes:
            if g.BioType != 'miRNA':
                tu = findinGene(g, start, end)
                results.append((name,g.Symbol,g.Location,g.BioType,tu,g.StableId))
                #print name,g.Symbol,tu
    if len(results)>0:
        results = pd.DataFrame(results,columns=['#miRNA','gene','location','biotype',
                                'tu','ensid'])
        return results
    return

#def getGeneProperty(genes, label):
#    x = [g[0].Symbol if len(g)>0 else np.nan for g in orthgenes]
#    return x

def get_mirna_orthologs(df, comp=None, ref='cow'):
    """Get all possible orthologs/conservation for miRNAs using ensembl"""

    if comp == None:
        comp = Compara(species, account=None, Release='79')
    results=[]
    for i, r in list(df.iterrows()):
        #base.RNAfold(r['consensus precursor sequence'], r['#miRNA']+'_'+ref)
        mature = r['consensus mature sequence'].replace('u','t').upper()
        seed = r['seed'].replace('u','t').upper()
        c,locs,strand = r['precursor coordinate'].split(':')
        start,end = locs.split('..')
        print (r['#miRNA'], seed, mature, locs, strand)
        coords = (c,int(start),int(end),strand)
        regions, aln = getSyntenicAlignment(comp, ref, coords, fname=r['#miRNA']+'.aln.fa')
        if aln == None:
            x=pd.DataFrame([r])
            a = getHostGenes(x)
            if a is not None:
                results.append(a)
            continue
        region = regions[0]
        a = base.cogentAlignment2DataFrame(aln.degap())
        a['#miRNA'] = r['#miRNA']
        a['ident'] = getIdentities(aln)
        print ('max identity: %s' %a.ident.max())
        a['seedcons'] = getSeqConservation(aln, seed)
        orthgenes = getGenesinRegion(region)

        a['gene'] = [g[0].Symbol if len(g)>0 else np.nan for g in orthgenes]
        a['gene_loc'] = [g[0].Location if len(g)>0 else np.nan for g in orthgenes]
        locs = getLocations(region)
        a['location'] = [':'.join(str(l).split(':')[2:]) for l in locs]
        a['biotype'] = [g[0].BioType if len(g)>0 else np.nan for g in orthgenes]
        a['ensid'] = [g[0].StableId if len(g)>0 else np.nan for g in orthgenes]
        #find where in gene the miRNA is, usually introns
        trpts = []
        for l,g in zip(locs,orthgenes):
            if len(g)==0:
                trpts.append(np.nan)
            else:
                tr = findinGene(g[0],l.Start,l.End)
                trpts.append(tr)
        a['tu'] = trpts

        #get RNAfold energy for each sequence
        a['energy'] = a.apply(lambda x : base.RNAfold(x.seq)[1],1)
        results.append(a)
        print (a)
        print ('--------------------------------------------------------')
    results = pd.concat(results).reset_index(drop=True)
    #results.drop(columns='seq')
    results.to_csv('novel_orthologs.csv')
    return

def get_locations(region):
    """Locations with region alignments"""

    l=[]
    for r in region.Members:
        try:
            loc = r.Location
            l.append(loc)
        except:
            continue
    return l

def get_seq_conservation(aln, seq):
    """Get position of a sub sequence and return position if it's
        conserved for each aligned species, -1 if not"""

    vals=[]
    for s in aln.Seqs:
        print (seq, s, str(s).find(seq))
        vals.append(str(s).find(seq))
    return vals

def get_identities(aln):
    """Identities for all seqs in a cogent alignment"""

    names=aln.Names
    ref=names[0]
    vals=[None]
    for n in names[1:]:
        new = aln.takeSeqs([ref, n])
        ident = round(len(new.filtered(lambda x: len(set(x)) == 1))/float(len(new)),3)
        vals.append(ident)
    return vals

def get_genes_in_region(region):
    """Get orthologous genes in all species in a syntenic region"""

    print ('checking genes in aligned species')
    orthologs=[]
    for r in region.Members:
        try:
            loc = r.Location
        except:
            continue
        sp = loc.Species
        coords = loc.CoordName,loc.Start,loc.End,loc.Strand
        #print sp, loc
        genes = list(r.genome.getFeatures(feature_types='gene',region=r))
        orthologs.append(genes)
    return orthologs

def get_ests(region):

    for r in region.Members:
        loc = r.Location
        sp = loc.Species
        ests = r.genome.getFeatures(feature_types='est', region=r)
        for est in ests:
            print (est)
    return

def get_sequence_from_location(species, coords):
    """Get sequence from a genomic location in an ensembl species genome."""

    from cogent.db.ensembl import HostAccount, Genome, Compara, Species
    genome = Genome(Species=species, Release='87', account=None)
    chrom,start,end,strand = coords
    #print coords
    r = genome.getRegion(CoordName=str(chrom), Start=start,End=end,Strand=strand)
    return r.Seq

def test():

    comp = Compara(species, account=account, Release=release)
    #print comp.method_species_links
    coords = [('15', 85273455, 85273507, '+'),('18',12423976,12424060,'+')]
    for c in coords:
        getSyntenicAlignment(comp, 'cow', c)
        #getAlignmentTree('ensembl_aln.fa')
    return
